'' =================================================================================================
''
''   File....... jm_sbus_pack_test.spin
''   Purpose....
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2021 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 19 MAR 2021
''
''   {$P1}
''
'' =================================================================================================


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' use 5MHz crystal

  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us


con { terminal }

  BR_TERM  = 115_200                                            ' terminal baud rate

  #0, T_PST, T_ANSI                                             ' terminal types

  T_TYPE = T_PST


con { fixed io pins }

  PGM_RX = 31  { I }                                            ' serial / programming
  PGM_TX = 30  { O }

  EE_SDA = 29  { I/O }                                          ' i2c / eeprom
  EE_SCL = 28  { I/O }


con { app io pins }


con

  #true,  ON, OFF
  #false, NO, YES


obj

' main                                                          ' * master Spin cog
  time : "jm_time_80"                                           '   timing and delays (80MHz)
  term : "jm_fullduplexserial"                                  ' * serial IO for terminal
  ansi : "jm_ansi"                                              '   ANSI terminal control sequences

' * uses cog when loaded


dat

' matching data sets for pack/unpack testing

  ChannelX      word    1027, 1024,  251, 1024,  240,  240,  240, 1807, 1024
                word    1807, 1024, 1024, 1024, 1024, 1024, 1024,    0,    0

  SBusX         byte    $0F, $03, $04, $E0, $3E, $00, $08, $0F, $78, $C0
                byte    $E3, $E1, $00, $7C, $38, $00, $01, $08, $40, $00
                byte    $02, $10, $80, $00, $00


var

  byte  sbus[25]


pub main | i, err, cog, et1, et2

  setup

  wait_for_terminal(true)
  time.pause(100)

  ' spin packing

  et1 := cnt
  pack(@ChannelX)
  et2 := cnt

  term.str(string("Spin", 13, 13))
  report

  term.dec((et2-et1)/US_001)
  term.str(string("us"))

  term.txn(13, 3)

  ' pasm packing

  p_chhub   := @ChannelX
  p_sbushub := @sbus
  p_flag    := @et1

  cog := cognew(@packfast, 0) + 1
  time.pause(5)
  cogstop(cog-1)

  term.str(string("PASM", 13, 13))
  report

  term.dec((et2-et1)/US_001)
  term.str(string("us"))

  repeat
    waitcnt(0)


pub report | i, err

  err := 0

  repeat i from 0 to 24
    term.fhex(sbus[i], 2)
    term.tx(" ")
    if ((i == 9) or (i == 19))
      term.tx(term#CR)

    if (sbus[i] <> SBusX[i])
      ++err

  term.fstr1(string("\r\rErrors: %d\r\r"), err)


pri pack(p_ch) | i, j, bc, tmp

  longfill(@i, 0, 4)
  i := 1

  sbus[0] := $0F                                                ' start byte

  repeat while (i < 23)                                         ' pack 22 bytes
    repeat while (bc < 8)                                       ' 8 bits / byte
      tmp |= (word[p_ch][j++] << bc)                            ' get value, position, add to tmp
      bc += 11                                                  ' bump bit count
    sbus[i++] := tmp                                            ' save byte
    tmp >>= 8                                                   ' remove this byte from tmp
    bc -= 8                                                     ' adjust bit count

  sbus[23] := 0                                                 ' clear flags

  if (word[p_ch][16] > 1023)                                    ' set flag if >= 50%
    sbus[23] |= %0001

  if (word[p_ch][17] > 1023)                                    ' set flag if >= 50%
    sbus[23] |= %0010

  sbus[24] := 0                                                 ' end byte


dat { pasm channel -> sbus packing }

                org

packfast        mov       t1, cnt                               ' mark start time
                wrlong    t1, p_flag


                mov       t1, #$0F                              ' set start byte
                mov       sbidx, #0
                wrbyte    t1, p_sbushub
                add       p_sbushub, #1
                mov       sbidx, #1                             ' pack 1..23

:loop1          rdword    t2, p_chhub                           ' get 11-bit channel value
                add       p_chhub, #2                           ' bump channel pointer

                mins      t2, #0                                ' keep in bounds
                maxs      t2, CH_MAX
                shl       t2, sbbc                              ' move into position
                or        sbuswork, t2                          ' add into workspace
                add       sbbc, #11                             ' update bits in sbuswork

:loop2          cmpsub    sbbc, #8                      wc      ' enough bits to store byte?
    if_nc       jmp       #:loop1                               '  if no, get next channel

                mov       t1, sbuswork                          ' copy for storage
                and       t1, #$FF                              ' truncate to 8 bits
                wrbyte    t1, p_sbushub
                add       p_sbushub, #1
                add       sbidx, #1                             ' increment storage index
                shr       sbuswork, #8                          ' remove last byte from workspace
                cmpsub    sbidx, #23                    wc      ' last byte of data?
    if_nc       jmp       #:loop2                               '  if no, pack next byte

                mov       t1, #0                                ' clear for digital channels

                rdword    t2, p_chhub                           ' read ch17
                add       p_chhub, #2
                cmp       t2, CH_MID                    wc
    if_a        or        t1, #%0001

                rdword    t2, p_chhub                           ' read ch18
                cmp       t2, CH_MID                    wc
    if_a        or        t1, #%0010

                mov       sbidx, #23                            ' write digital channels
                wrbyte    t1, p_sbushub
                add       p_sbushub, #1

                mov       t1, #$00                              ' set end byte
                mov       sbidx, #24
                wrbyte    t1, p_sbushub
                add       p_sbushub, #1

                mov       t1, cnt                               ' mark end time
                add       p_flag, #4
                wrlong    t1, p_flag

                jmp       #$


p_chhub         long      0-0                                   ' updat before launching cog
p_sbushub       long      0-0
p_flag          long      0-0

sbidx           long      0
sbuswork        long      0
sbbc            long      0

CH_MAX          long      $07FF
CH_MID          long      $03FF

t1              long      0
t2              long      0


pub setup

'' Setup IO and objects for application

  time.start                                                    ' setup timing & delays

  term.start(PGM_RX, PGM_TX, %0000, BR_TERM)                    ' start serial for terminal *


pub wait_for_terminal(clear)

'' Wait for terminal to be open and key pressed

  term.rxflush
  term.rx
  if (clear)
    clear_screen


pub clear_screen

  if (T_TYPE == T_PST)
    term.tx(term#CLS)
  else
    term.str(ansi.hide_cursor)
    term.str(ansi.home)
    term.str(ansi.cls)


con { license }

{{

  MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

}}